(* Time-stamp: <modified the 13/02/2013 (at 08:51) by Erwan Jahier> *)

(** Utilities for managing node environements (items tables)  *)


type t = {
  id2const : Ident.idref -> Lxm.t -> Lic.const;
  id2type  : Ident.idref -> Lxm.t -> Lic.type_;
  id2node  : Ident.idref -> Lic.static_arg list -> Lxm.t -> Lic.node_exp;

  id2var   : Ident.t -> Lxm.t -> Lic.var_info;
  global_symbols  : AstTabSymbol.t;
}

type local_env = {
  lenv_node_key : Lic.node_key ;
(*   lenv_globals : pack_env ; *)
  lenv_types : (Ident.t, Lic.type_) Hashtbl.t ;
  lenv_const : (Ident.t, Lic.const) Hashtbl.t ; 
  lenv_nodes : (Ident.t, Lic.node_key) Hashtbl.t ; 
  lenv_vars  : (Ident.t, Lic.var_info) Hashtbl.t ; 
}


let (make_local_env : Lic.node_key -> local_env) =
  fun nk ->
    let res =
      {
        lenv_node_key = nk;
        lenv_types = Hashtbl.create 0;
        lenv_const = Hashtbl.create 0;
        lenv_nodes = Hashtbl.create 0;
        lenv_vars  = Hashtbl.create 0;
      }
    in
      (* fill tables using static arg info *)
      List.iter
        (function
           | Lic.ConstStaticArgLic(id,ce) -> Hashtbl.add res.lenv_const id ce
           | Lic.TypeStaticArgLic(id,te)  -> Hashtbl.add res.lenv_types id te
           | Lic.NodeStaticArgLic(id, nk) -> Hashtbl.add res.lenv_nodes id nk
        )
        (snd nk);
      
      res

let dump_local_env oc e =
   let pt i t = Printf.fprintf oc "#    type %s = %s\n" i (Lic.string_of_type t) in
   Hashtbl.iter pt e.lenv_types;
   let pc i t = Printf.fprintf oc "#    const %s = %s\n" i (Lic.string_of_const t) in
   Hashtbl.iter pc e.lenv_const;
   (* let pn i (n,_,_) = Printf.fprintf oc "#    node %s = %s\n" i (string_of_node_key n) in *)
   let pn i nk = Printf.fprintf oc "#    node %s = %s\n" i (Lic.string_of_node_key nk) in
   Hashtbl.iter pn e.lenv_nodes;


(* Grouping those 2 ones is sometimes useful *)
type node_env = {
  local : local_env;
  global: t;
}

let (lookup_type: local_env -> Ident.idref -> Lxm.t -> Lic.type_) =
  fun env id lxm ->
    Hashtbl.find env.lenv_types (Ident.of_idref id)

let (lookup_node : local_env -> Ident.idref -> Lxm.t -> Lic.node_key) =
  fun env id lxm ->
    Hashtbl.find env.lenv_nodes (Ident.of_idref id)
    
let (lookup_const: local_env -> Ident.idref -> Lxm.t -> Lic.const) =
  fun env id lmx ->  
    Hashtbl.find env.lenv_const (Ident.of_idref id)
          
let (lookup_var: local_env -> Ident.t -> Lxm.t -> Lic.var_info) = 
  fun env id lmx ->
    Hashtbl.find env.lenv_vars id



let node_exp_of_node_key
    (id_solver: t) (node_key: Lic.node_key) (lxm : Lxm.t)
    : Lic.node_exp =
  let (id, sargs) = node_key in
  id_solver.id2node (Ident.idref_of_long id) sargs lxm

let var_info_of_ident 
    (id_solver: t) (id: Ident.t) (lxm : Lxm.t)
    : Lic.var_info =
  id_solver.id2var id lxm

let const_eff_of_item_key
    (id_solver: t) (id: Lic.item_key) (lxm : Lxm.t) 
    : Lic.const =
  id_solver.id2const (Ident.idref_of_long id) lxm



