

module ItemKeyMap = struct
	include Map.Make (
		struct
			type t = Lic.item_key
			let compare = compare
		end
	)
	let dummy () = "dummy item: add things below to complete the module"
end

module NodeKeyMap = struct
	include Map.Make (
		struct
			type t = Lic.node_key
			let compare = compare
		end
	)
	let dummy () = "dummy item: add things below to complete the module"
end

type t = {
  types  : Lic.type_ ItemKeyMap.t;
  consts : Lic.const ItemKeyMap.t;
  nodes  : Lic.node_exp NodeKeyMap.t;
  (* consts : (Lic.item_key, Lic.const    Lic.check_flag) Hashtbl.t; *)
  (* nodes  : (Lic.node_key, Lic.node_exp Lic.check_flag) Hashtbl.t; *)
}

let empty = {
   types = ItemKeyMap.empty;
   consts = ItemKeyMap.empty;
   nodes = NodeKeyMap.empty
}

(* KoKêterie : pour changer ds suffixes 
   numériques, produit :
   a, b, c, d, e, aa, ab, ac, ad, ae, etc ...
Pour tester :
for i = 0 to 42 do
Printf.printf "%3d -> %s\n" i (pretty_sfx i)
done
*)
let rec pretty_sfx i =
   if i = 0 then ""
   else
      (pretty_sfx ((i-1)/5))^(Char.escaped (char_of_int (97 + (i-1) mod 5)))

(** CREER DES IDENTS TOUT FRAIS *)
let fresh_type_id this pname pfx =
   let rec fresh x =
      let id = Printf.sprintf "%s%s" pfx (pretty_sfx x) in
      let res = Ident.make_long pname id in
      if ItemKeyMap.mem res this.types then fresh (x+1)
      else res
   in
   fresh 0

(** RECHERCHE *)
let find_type  this k = try Some(ItemKeyMap.find k this.types ) with Not_found -> None
let find_const this k = try Some(ItemKeyMap.find k this.consts) with Not_found -> None
let find_node  this k = try Some(NodeKeyMap.find k this.nodes ) with Not_found -> None
   
let node_exists this k = NodeKeyMap.mem k this.nodes


let (find_var : Ident.t -> Lic.node_exp -> Lic.var_info option) = 
  fun id ne -> 
    let name_matches vi = vi.Lic.var_name_eff = id in
    try Some (List.find name_matches ne.Lic.inlist_eff) with Not_found -> 
    try Some (List.find name_matches ne.Lic.outlist_eff) with Not_found -> 
    match ne.Lic.loclist_eff with
      | None -> None
      | Some vil -> 
        try Some (List.find name_matches vil)
        with Not_found -> None


(** PARCOURS *)
let fold_consts (f: Lic.item_key -> Lic.const -> 'a -> 'a) (this:t) (accin:'a) : 'a =
   ItemKeyMap.fold f this.consts accin
let fold_types (f: Lic.item_key -> Lic.type_ -> 'a -> 'a) (this:t) (accin:'a) : 'a =
   ItemKeyMap.fold f this.types accin
let fold_nodes (f: Lic.node_key -> Lic.node_exp -> 'a -> 'a) (this:t) (accin:'a) : 'a =
   NodeKeyMap.fold f this.nodes accin

let list_nodes t = fold_nodes (fun k e acc -> (k,e)::acc) t []

let choose_node t = 
  (* since ocaml 3.12.0 only... NodeKeyMap.choose *)
  match list_nodes t with
    | n::_ -> Some n
    | [] -> None


let iter_consts (f: Lic.item_key -> Lic.const -> unit) (this:t) : unit =
   ItemKeyMap.iter f this.consts
let iter_types (f: Lic.item_key -> Lic.type_ -> unit) (this:t) : unit = 
   ItemKeyMap.iter f this.types
let iter_nodes (f: Lic.node_key -> Lic.node_exp -> unit) (this:t) : unit =
   NodeKeyMap.iter f this.nodes



let add_type (k:Lic.item_key) (v:Lic.type_) (prg:t) : t =
   { prg with types = ItemKeyMap.add k v prg.types }

let add_const (k:Lic.item_key) (v:Lic.const) (prg:t) : t =
   { prg with consts = ItemKeyMap.add k v prg.consts }

let add_node (k:Lic.node_key) (v:Lic.node_exp) (prg:t) : t =
  Verbose.exe ~level:3 (fun () ->
    Printf.printf  "## LicPrg.add_node %s\n" (LicDump.string_of_node_key_rec k));
   { prg with nodes = NodeKeyMap.add k v prg.nodes }

exception Print_me of Lic.node_exp
let to_file (oc: out_channel) (this:t) (main_node: Ident.idref option) = 
  Verbose.dump_entete oc;
  (* On imprime dans l'ordre du iter, donc pas terrible ???
  *)
  ItemKeyMap.iter
    (fun tn te -> 
      if (not !Global.ec || Lic.is_extern_type te) then 
        output_string !Global.oc (LicDump.type_decl tn te)
    )
    this.types;

  (* for generating lv4 or ec compatible code, enum types are
     translated into an extern type + an extern const per enums.

     For instance, 

     type color1 = enum { blue, white, black };
     type color2 = enum { green, orange, yellow };

     will be translated into 

     type color1;
     type color2;
     const orange:color2;
     const green:color2;
     const black:color1;
     const yellow:color2;
     const blue:color1;
     const white:color1;

  *)
  if !Global.expand_enums then (
    let const_list = 
      ItemKeyMap.fold
        (fun tn te acc -> 
          match te with
            | Lic.Enum_type_eff(long, longl) ->  
              output_string !Global.oc (LicDump.type_decl long (Lic.External_type_eff long));
              List.rev_append (List.map (fun x -> long,x) longl) acc
            | _ -> acc
        )
        this.types
        []
    in
    List.iter
      (fun (t,elt) -> 
        let const = Lic.Extern_const_eff (elt, Lic.External_type_eff t) in
        output_string !Global.oc (LicDump.const_decl elt const))
      const_list;
  );
  ItemKeyMap.iter
    (fun cn ce -> 
      if (not !Global.ec || Lic.is_extern_const ce) then
        output_string !Global.oc (LicDump.const_decl cn ce)
    )
    this.consts  ;
  if !Global.ec then (
    (* If no node is set a top-level, the compiler will compile every node. But the
       ec format only accepts one node (and no type nor const) 
       Hence we print the first one (if no main node is set).
    *)

    try 
      NodeKeyMap.iter 
        (fun (key,_) nexp -> (
          match main_node with
            | Some { Ident.id_pack = None ; Ident.id_id= name } -> 
              if Ident.of_long key = name then raise (Print_me nexp)
            | Some idref -> 
              if Ident.long_of_idref idref = key then raise (Print_me nexp)
            | None -> (
              match nexp.Lic.node_key_eff, nexp.Lic.def_eff with
                (* only user or extern nodes with a body makes valid ec node *)
                | _, Lic.BodyLic _ -> raise (Print_me nexp)
                | (("Lustre",_),[]), Lic.ExternLic -> ()
                | _, Lic.ExternLic ->
                  (* we need to declare all extern nodes (do we?) *)
                  output_string !Global.oc (LicDump.node_of_node_exp_eff nexp);
                  flush !Global.oc;
                | _  -> ()
            )
         )
        )
        this.nodes        
    with Print_me nexp -> 
      output_string !Global.oc (LicDump.node_of_node_exp_eff nexp);
      flush !Global.oc;
  ) else (
    (* Pour les noeuds, pas sur que ça marche tant qu'on n'a
       pas séparés les transformations  source_to_source du LicTab:
       en cas d'expansion, il y avait cette remarque :
       nb: we print res_struct, but do not return it from
       node_check, because the structure and array expansion
       modify (instanciate) the node profiles.

       On n'affiche PAS les extern Lustre::...
    *)

    NodeKeyMap.iter (
      fun _ nexp ->
        match nexp.Lic.node_key_eff with
          (* inutile d'écrire les noeuds predefs *)
          | (("Lustre",_),[]) -> ()
          | _ -> output_string !Global.oc (LicDump.node_of_node_exp_eff nexp)
    )
      this.nodes
  )

(* GENERATEUR DE NOM DE VARIABLES *)
type id_generator = string -> string

let fresh_var_id_generator : t -> Lic.node_exp -> id_generator =
  fun prg ne ->
    let cpt = ref 0 in
    let forbidden = Hashtbl.create 100 in
    let _ = iter_consts (fun i c -> match c with
      | Lic.Extern_const_eff (s,_)
      | Lic.Abstract_const_eff (s,_,_,_)
      | Lic.Enum_const_eff   (s,_) -> Hashtbl.add forbidden (snd s) ()
      | _ -> ()
    ) prg in
    let dovar vi = Hashtbl.add forbidden vi.Lic.var_name_eff () in
    let _ = List.iter dovar ne.Lic.inlist_eff in
    let _ = List.iter dovar ne.Lic.outlist_eff in
    let _ = match ne.Lic.loclist_eff with
      | Some l -> List.iter dovar l | None -> ()
    in
    let rec dogen (pfx: string) : string =
      let id = Printf.sprintf "%s%02d" pfx !cpt in
      incr cpt;
      try (
        let _ = Hashtbl.find forbidden id in
        dogen pfx
      ) with Not_found -> (
        Hashtbl.add forbidden id ();
        id 
      )
    in
    dogen
