(*
Le manager d'argument adapt de celui de lutin, plus joli
N.B. solution un peu batarde : les options sont stockes, comme avant, dans Global,
du coup, le fait de rendre un type "t" est une koketerie !
*)
open Version
open Verbose
open Arg

let tool_name = Version.tool
let usage_msg =  "usage: "^tool_name^" [options] <file> | "^tool_name^" -help"

type t = {
  mutable _opts : (string * Arg.spec * string) list; (* classical Arg option tab used by Arg.parse *)
  mutable _user_man  : (string * string list) list; (* ad hoc tab for pretty prtting usage *)
  mutable _hidden_man: (string * string list) list; (* ad hoc tab for pretty prtting usage *) 
  mutable _others: string list;
  mutable _margin : int;

}

let (make_opt : unit -> t) = 
  fun () -> 
  {
   _opts = [];        
   _user_man  = [];   
   _hidden_man  = []; 
   _others = [];
   _margin = 12;
}

(* all unrecognized options are accumulated *)
let (add_other : t -> string -> unit) =
  fun opt s -> 
    opt._others <- s::opt._others

let pspec os opt (c, ml) = (
  let (m1, oth) = match ml with
	 |	h::t -> (h,t)
	 |	_ -> ("",[])
  in
  let t2 = String.make opt._margin ' ' in
  let cl = String.length c in
  let t1 = if (cl < opt._margin ) then
	 String.make (opt._margin - cl) ' '
  else
	 "\n"^t2
  in
	 Printf.fprintf os "%s%s%s" c t1 m1;
	 List.iter (function x -> Printf.fprintf os "\n%s%s" t2 x) oth ;
	 Printf.fprintf os "\n" ;
)

let usage os opt = (
	let l = List.rev opt._user_man in
	Printf.fprintf os "%s\n\n" usage_msg;
	List.iter (pspec os opt) l
)
let help opt ()= (
	usage stdout opt;
	exit 0
)
let full_usage os opt = (
	Printf.fprintf os "%s\n" usage_msg;
(* 	let l = List.rev opt._user_man in *)
(* 	List.iter (pspec os opt) l; *)
	let l = List.rev (opt._hidden_man) in
	List.iter (pspec os opt) l
)
let full_help opt ()= (
	full_usage stdout opt;
	exit 0
)

let unexpected s opt = (
	prerr_string ("unexpected argument \""^s^"\"");
	prerr_newline ();
	usage stderr opt;
	exit 1
)
let file_notfound f opt = (
	prerr_string ("File not found: \""^f^"\"");
	prerr_newline ();
	usage stderr opt;
	exit 1
)

let (mkopt : t -> string list -> ?hide:bool -> ?arg:string -> Arg.spec -> string list -> unit) =
  fun opt ol ?(hide=false) ?(arg="") se ml ->
    let treto o = opt._opts <- (o, se, "")::opt._opts in
	   List.iter treto ol ;
	   let col1 = (String.concat ", " ol)^arg in
	     if hide 
        then opt._hidden_man <- (col1, ml)::opt._hidden_man
	     else opt._user_man   <- (col1, ml)::opt._user_man
          (*
	         let tabs = String.make (col - (String.length o) - (String.length arg)) ' ' in
	       (* (o, se, arg^tabs^m) *)
	         (o, se, arg^"\n     "^m)
          *)

(* utils *)
let set_v4_options () =
  Global.lv4 := true;
  Global.inline_iterator := true;
  Global.expand_arrays := true;
  Global.expand_enums := true

let set_ec_options () =
  set_v4_options ();
  Global.ec := true;
  Global.no_prefix := true;
  Global.expand_nodes := true

(*** USER OPTIONS TAB **)
let mkoptab (opt:t) : unit = (
    mkopt opt
      ["-n";"-node"]
      ~arg:" <string>"
      (Arg.String(function x -> 
         Global.main_node := x;
         Global.compile_all_items := false))
      ["Set the main node (all items are compiled if unset)"]
    ;
    mkopt opt
      ["-o";"--output-file"]
      ~arg:" <string>"
      (Arg.String(function x -> 
         Global.outfile := x))
      ["Set the output file name"]
    ;
    mkopt opt
      ["-exec"]
      (Arg.Unit (fun _ -> Global.exec := true))
      ["interpret the program using RIF conventions for I/O (experimental)."]
    ;

    mkopt opt
      ["-knc"; "--keep-nested-calls"]
      (Arg.Unit (fun _ -> Global.one_op_per_equation := false))
      ["Keep nested calls (inhibited by -en). By default, only one node per equation is generated."]
    ;
    mkopt opt
      ["-ei"; "--expand-iterators"]
      (Arg.Unit (fun _ -> Global.inline_iterator := true))
      ["Expand array iterators (i.e., generate iterator-free code)."]
    ;
    mkopt opt
      ["-ee"; "--expand-enums"]
      (Arg.Unit (fun _ -> Global.expand_enums := true))
      [" Translate enums using extern types and consts (for lv4 and ec)."]
    ;
    mkopt opt
      ["-esa"; "--expand-structs-and-arrays"]
      (Arg.Unit (fun _ ->
         Global.expand_arrays := true;
         Global.inline_iterator := true))
      ["Expand structures and arrays using as many variables as necessary (automatically impose '-ei')"]
    ;
    mkopt opt
      ["-en"; "--expand-nodes"]
      (Arg.Unit (fun _ -> Global.expand_nodes := true))
      ["Expand the main node (use the first node if no one is specified)."]
    ;
    mkopt opt
      ["-den"; "--do_not-expand-nodes"]
      ~arg:" <string>"
      (Arg.String (fun str ->
         Global.dont_expand_nodes := str::!Global.dont_expand_nodes
      ))
      ["Do not expand the specified node (meaningful with -en only of course)."]
    ;
    mkopt opt
      ["-lv4"; "--lustre-v4"]
      (Arg.Unit (fun _ -> set_v4_options ()))
      ["Use Lustre V4 syntax (automatically impose '-ei -ee -esa')."]
    ;
    mkopt opt
      ["-ec"; "--expanded-code"]
      (Arg.Unit (fun _ -> set_ec_options ()))
      ["Generate ec (actually just an alias for '-en -lv4 --no-prefix')."]
    ;
    mkopt opt
      ["-np"; "--no-prefix"]
      (Arg.Set Global.no_prefix)
      ["Do not prefix variable names by their module (beware: variable names may clash with this option)."]
    ;

    mkopt opt
      ["-version"; "--version"]
      (Arg.Unit(function _ -> Printf.fprintf stdout "%s\n" Version.str; exit 0))
      ["Print the current version and exit"]
    ;
    (* verbose *)
    mkopt opt
      ["-v"; "--verbose"]
      (Arg.Unit(function _ -> Verbose.on () ))
      ["Set the verbose level to 1"]
    ;
    mkopt opt
      ["-vl"]
      ~arg:" <int>"
      (Arg.Int(function i -> Verbose.set i))
      ["Set the verbose level"]
    ;
    mkopt opt
      ["-h";"-help";"--help"]
      (Arg.Unit (help opt))
      ["Display this message."]
    ;
    (* to show Hidden opt *)
    mkopt opt
      ["-more"]
      (* (Arg.Unit(fun _ -> opt._see_all_options <- true)) *)
      (Arg.Unit (full_help opt))
      ["Show hidden options (for dev purposes)"];
    (* HIDDEN *)

    (* test lexical *)
    mkopt opt ~hide:true
      ["-tlex"; "--test-lexer"]
      (Arg.Set Global.tlex)
      ["Test the lexical analysis"]
    ;
    (* test syntaxique
    mkopt opt ~hide:true
      ["-tparse"]
      (Arg.Unit(function _ -> opt._gen_mode <- GenLuc ; opt._test_parse <- true ; ()))
      ["Test the syntactic analysis"]
    ;
    *)
    mkopt opt ~hide:true
      ["-unit"]
      (Arg.Set Global.run_unit_test)
      ["Run embedded unit tests"]
    ;
    mkopt opt ~hide:true
      ["--nonreg-test"]
      (Arg.Set Global.nonreg_test)
      ["Avoid printing full path error msgs to ease non-reg test decision"]
    ;
    (* misc debug flag *)
    mkopt opt ~hide:true
      ["-dbg"; "--debug"]	
      (Arg.Symbol
         ( Verbose.flag_list (), fun s -> let f = Verbose.get_flag s in Verbose.set_flag f))
      [ "<dbg_flag>"; 
        "Possible dbg_flag are: " ^(String.concat ", " (Verbose.flag_list())) ]
)


let first_line b = (
	try (
		let f = String.index b '\n' in
		String.sub b 0 f
	) with Not_found -> b
)


let current = ref 0;;

(* La ``mthode'' principale *)
let parse argv = (
  let opt = make_opt() in
  let save_current = !current in
    try (
      mkoptab opt;
	   Arg.parse_argv ~current:current argv opt._opts (add_other opt) usage_msg;
      (List.iter 
         (fun f -> 
            if (String.sub f 0 1 = "-") then
              unexpected f opt
            else if not (Sys.file_exists f) then
              file_notfound f opt
            else ()
         ) 
         opt._others
      );
      Global.infiles := (List.rev opt._others);
      current := save_current;
      opt
    ) with
	     (* only 1rst line is interesting ! *)
	   | Bad  msg -> Printf.fprintf stderr "%s\n" (first_line msg); usage stderr opt; exit 2; 
	   | Help msg -> help opt ();
)

