(** Time-stamp: <modified the 29/03/2013 (at 16:05) by Erwan Jahier> *)


open Soc


(** Donne toute les mthodes d'un composant.

    C'est la liste des mthodes du composant, et la mthode d'initialisation le
    cas chant. *)
let get_all_methods: Soc.t -> step_method list = fun c ->
  c.step
(* DELETE ME !! *)
(*   match c.init with *)
(*     | None   -> c.step *)
(*     | Some m -> m :: c.step *)

(** Fonctions de reprsentation des objets LOC. *)

(** Aliases *)
let str_ff       = Format.str_formatter
let flush_str_ff = Format.flush_str_formatter
let fprintf      = Format.fprintf

(** Encapsule l'appel  une fonction avec formatter pour sortir une string. *)
let call_fun_ff: ((Format.formatter -> unit) -> string) = fun f ->
  let b = Buffer.create 50 in
  let ff = Format.formatter_of_buffer b in
    f ff;
    Format.pp_print_flush ff ();
    let s = Buffer.contents b in
      Buffer.reset b;
      s

(* Type *)
let rec string_of_type_ref_ff: (Soc.var_type -> Format.formatter -> unit) = fun v ff -> 
  let str =
    match v with
      | Soc.Bool -> "bool"
      | Soc.Int -> "int"
      | Soc.Real-> "real"
      | Soc.Extern s -> s ^ "(*extern*)"
      | Soc.Enum  (s, sl) -> "enum " ^ s ^ " {" ^ (String.concat ", " sl) ^ "}"
      | Soc.Struct (sid,_) -> sid ^ "(*struct*)"
      | Soc.Array (ty, sz) -> Printf.sprintf "%s^%d" (string_of_type_ref ty) sz 
      | Soc.Alpha nb ->
        (* On gnre des "types"  la Caml : 'a, 'b, 'c, etc. *)
        let a_value = Char.code('a') in
        let z_value = Char.code('z') in
        let str =
          if (nb >= 0 && nb <= (z_value - a_value)) then
            ("'" ^ (Char.escaped (Char.chr(a_value + nb))))
          else
            ("'a" ^ (string_of_int nb))
        in
          str
  in
    fprintf ff "%s" str 

and string_of_type_ref: (Soc.var_type -> string) = fun v ->
  call_fun_ff (string_of_type_ref_ff v)


(* Cl de composant *)
let string_of_soc_key_ff: (Soc.key -> Format.formatter -> unit) = 
  fun (id, types, si_opt) ff ->
    (match types with
      | [] -> fprintf ff "%s" id
      | _  -> fprintf ff "%s:%s" id 
          (String.concat " -> " (List.map string_of_type_ref types)));
    (match si_opt with
       | None -> ()
       | Some(f,l,step) -> fprintf ff "[%d .. %d step %d]"  f l step)

let string_of_soc_key: (Soc.key -> string) = fun v ->
  call_fun_ff (string_of_soc_key_ff v)


(* Variable *)
let string_of_var_ff: (Soc.var -> Format.formatter -> unit) = fun (id, type_) ff ->
  fprintf ff "%s: %s" id (string_of_type_ref type_)

let string_of_var: (Soc.var -> string) = fun v ->
  call_fun_ff (string_of_var_ff v)


let string_of_instance_ff: (instance -> Format.formatter -> unit) = 
  fun (name,sk) ff ->
    fprintf ff "%s" name

let string_of_instance: (instance -> string) = fun (name,sk) -> name
  

(* Opration *)
let string_of_operation_ff: (atomic_operation -> Format.formatter -> unit) = fun v ff -> match v with
  | Assign          -> () (* On suppose qu'il est dj affich dans string_of_gao *)
  | Method((n, sk),sname) -> fprintf ff "%s.%s" n sname
  | Procedure(proc)    -> fprintf ff "%s" (string_of_soc_key proc)

let string_of_operation: (atomic_operation -> string) = fun v ->
  call_fun_ff (string_of_operation_ff v)


(* Filtre d'accs *)
let rec string_of_filter_ff: (Soc.var_expr -> Format.formatter -> unit) = 
  fun v ff -> match v with
    | Const(id, _) 
    | Var (id,_)   -> fprintf ff "%s" id
    | Field(f, id,_) -> string_of_filter_ff f ff; fprintf ff ".%s" id
    | Index(f, index,_) -> string_of_filter_ff f ff; fprintf ff "[%d]" index
          
let string_of_filter: (Soc.var_expr -> string) = fun v ->
  call_fun_ff (string_of_filter_ff v)

(* Code *)
let rec string_of_gao_ff: (gao -> Format.formatter -> unit) = fun v ff -> match v with
  | Case (ck, cases) ->
      let string_of_case: (ident * gao list -> unit) = fun (id, c) ->
        fprintf ff "@[case %s:@[" id;
        string_of_gaos_list_ff c ff;
        fprintf ff "@]@]"
      in
        fprintf ff "switch(%s) {" ck;
        List.iter string_of_case cases;
        fprintf ff "}"

  | Call(dests, op, srcs) ->
      let _ =
        match dests with
          | [] -> () (* pas de destinations, on affiche pas de "=" *)
          | _  -> 
              let dests = String.concat ", " (List.map string_of_filter dests) in
                fprintf ff "%s = " dests
      in
      let srcs  = String.concat ", " (List.map string_of_filter srcs) in
        string_of_operation_ff op ff;
        fprintf ff "(%s)" srcs;

and string_of_gaos_list_ff: (gao list -> Format.formatter -> unit) = fun gaos ff ->
  List.iter (
    fun c ->
      fprintf ff "@[";
      string_of_gao_ff c ff;
      fprintf ff ";@]@,"
  ) gaos

let string_of_gao: (gao -> string) = fun v ->
  call_fun_ff (string_of_gao_ff v)

let string_of_gaos_list: (gao list -> string) = fun v ->
  call_fun_ff (string_of_gaos_list_ff v)


(* Profil de mthode *)
let string_interface_of_method_ff: (Soc.t -> step_method -> Format.formatter -> unit) = fun c m ff ->
  let string_var_from_index: (Soc.var list -> int -> string) = fun vl i ->
    if i < 0 then "mem" else
    string_of_var (List.nth vl i)
  in
    fprintf ff "%s(%s) -> (%s)"
      m.name
      (String.concat "; " (List.map (string_var_from_index (fst c.profile)) m.idx_ins))
      (String.concat "; " (List.map (string_var_from_index (snd c.profile)) m.idx_outs))


let string_interface_of_method: (Soc.t -> step_method -> string) = fun c m ->
  call_fun_ff (string_interface_of_method_ff c m)


(* Mthode complte *)
let string_of_method_ff: (Soc.t -> step_method -> Format.formatter -> unit) = fun c m ff ->

  fprintf ff "@[<v>@[<v 2>";
  string_interface_of_method_ff c m ff;

  match m.impl with
    | Predef -> fprintf ff "@]@]"
    | Iterator _ -> assert false (* todo *)
    | Gaol (locals, gaos) ->
          fprintf ff ": {@;";
          fprintf ff "@[<v>-- locals vars@;";
          List.iter (
            fun v ->
              string_of_var_ff v ff;
              fprintf ff ";@,";
          ) locals;
          fprintf ff "@]@;@[<v>-- code@;";
          string_of_gaos_list_ff gaos ff;
          fprintf ff "@]@]@ }@]"

let string_of_method: (Soc.t -> step_method -> string) = fun c m ->
  call_fun_ff (string_of_method_ff c m)


(* Ordre des mthodes *)
let string_of_precedence_ff: (string * string list -> Format.formatter -> unit) = fun (m, needs) ff ->
  fprintf ff "%s < [%s]" m (String.concat "; " needs)

let string_of_precedence: (string * string list -> string) = fun v ->
  call_fun_ff (string_of_precedence_ff v)


(** Profile d'un composant *)
let string_of_profile_ff: Soc.var list * Soc.var list -> Format.formatter -> unit = fun (ins, outs) ff ->
  fprintf ff "profile: @[(%s) ->@ (%s)@]"
    (String.concat "; " (List.map string_of_var ins))
    (String.concat "; " (List.map string_of_var outs))

let string_of_profile: Soc.var list * Soc.var list -> string = fun profile ->
  call_fun_ff (string_of_profile_ff profile)


(* Convertion des lments d'un composant *)
(* Convertion du profil ... *)
let string_of_soc_profile_ff: (Soc.t -> Format.formatter -> unit) = fun comp ff ->
  string_of_profile_ff comp.profile ff

let string_of_soc_profile: (Soc.t -> string) = fun comp ->
  call_fun_ff (string_of_soc_profile_ff comp)

(* ... des contraintes *)
let string_of_comp_constraints_ff: (Soc.t -> Format.formatter -> unit) = fun comp ff ->
  fprintf ff "constraints: @[";
  match comp.precedences with
    | [] -> fprintf ff "[]@]"
    | _ ->
        fprintf ff "%s@]"
          (String.concat "; " (List.map string_of_precedence comp.precedences))

let string_of_soc_factory_ff: (
  Soc.t -> Format.formatter ->
  (Soc.t -> step_method -> Format.formatter -> unit) -> (* Formatage des mthodes *)
  (instance -> Format.formatter -> unit) option -> (* Formatage des mmoires *)
  unit
) = fun comp ff format_meth format_mem ->
  let display_mem () = 
    match format_mem with
      | None -> ()
      | Some f -> (
        fprintf ff "@[<v 2>instances:@,";
        List.iter (
          fun m ->
            f m ff;
            fprintf ff ";@,"
        ) comp.instances
      )
  in
  let display_init () = ()
(* DELETE ME !! *)
(*     match comp.init with *)
(*       | None -> fprintf ff "@[<v 2>init: -@]" *)
(*       | Some m -> ( *)
(*         fprintf ff "@[<v 2>init:@,"; *)
(*         format_meth comp m ff; *)
(*         fprintf ff "@]" *)
(*       ) *)
  in

  fprintf ff "@[<v>@[<v 2>soc ";
  string_of_soc_key_ff comp.key ff;
  fprintf ff ":@,@[<v>";

  string_of_soc_profile_ff comp ff;
  fprintf ff "@]@,@[<v>";

  display_mem();
  fprintf ff "@]@,@[<v>";

  string_of_comp_constraints_ff comp ff;
  fprintf ff "@]@,@[<v>";

  display_init();
  fprintf ff "@]@,@[<v>";

  fprintf ff "@[<v 2>steps:@,";
  List.iter (
    fun s ->
      fprintf ff "@[";
      format_meth comp s ff;
      fprintf ff "@]@,"
  ) comp.step;

  fprintf ff "@]@]@]@]@."


(* Interface d'un composant *)
let string_interface_of_soc_ff: (Soc.t -> Format.formatter -> unit) = fun comp ff ->
  string_of_soc_factory_ff comp ff
    string_interface_of_method_ff
    None

let string_interface_of_soc: (Soc.t -> string) = fun v ->
    call_fun_ff (string_interface_of_soc_ff v)


(* Composant complet *)
let string_of_soc_ff: (Soc.t -> Format.formatter -> unit) = fun comp ff ->
  string_of_soc_factory_ff comp ff
    string_of_method_ff
    (Some string_of_instance_ff)

let string_of_soc: (Soc.t -> string) = fun v ->
    call_fun_ff (string_of_soc_ff v)


let dump_entete oc = 
  let time = Unix.localtime (Unix.time ()) in
  let sys_call, _ = Array.fold_left 
    (fun (acc,i) x -> 
       if 70 < i + (String.length x) then 
	 acc ^ "\n--\t\t" ^ x, String.length ("\n--\t\t" ^ x)
       else 
	 acc ^ " " ^ x , (i+1+(String.length x))
    )
    ("",0) 
    Sys.argv
  and 
      date = ( 
	(string_of_int time.Unix.tm_mday) ^ "/" ^
	  (string_of_int (time.Unix.tm_mon+1)) ^  "/" ^
	  (string_of_int (1900+time.Unix.tm_year))
      )
  and time_str = (
    (string_of_int time.Unix.tm_hour) ^  ":" ^
      (if time.Unix.tm_min < 10 then "0" else "") ^
      (string_of_int time.Unix.tm_min) ^   ":" ^
      (if time.Unix.tm_sec < 10 then "0" else "") ^
      (string_of_int time.Unix.tm_sec) 
  )
    (*   and user = Unix.getlogin () *)
  and hostname = Unix.gethostname ()
  in
    output_string oc
      ("-- This file was generated by "^Sys.argv.(0)^" version " ^ Version.str ^ 
         ".\n--\t" ^ sys_call ^ "
-- on " ^ hostname ^ 
         (* 	 "by "^ user ^ *)
         " the " ^ date ^ " at " ^ time_str ^ "\n\n");
    flush oc


let output: (bool -> string -> Soc.t list -> unit) = 
  fun no_header pkg_name socs ->
    let header = "Package '" ^ pkg_name ^ "' :" in
    let deco   = (String.make (String.length header) '=') in

      if no_header then () else dump_entete stdout ;
      print_string (deco ^ "\n" ^ header ^ "\n" ^ deco ^ "\n" ^ "\n");
      print_string (
        String.concat "\n\n" (List.map string_of_soc socs)
      );
      flush stdout



let (find : Lxm.t -> Soc.key -> Soc.tbl -> Soc.t) =
  fun lxm sk soc_tbl -> 
    try SocMap.find sk soc_tbl 
    with Not_found ->
      let str_of_key = string_of_soc_key in
      let nodes = SocMap.fold (fun sk _ acc -> (str_of_key sk)::acc) soc_tbl [] in
      let nodes_str = String.concat "\n\t" nodes in
      let msg = Printf.sprintf
        ("Cannot find a soc for the node %s. \n Available nodes are: \n\t%s\n") 
        (str_of_key sk)
        nodes_str
      in
      raise (Errors.Compile_error(lxm,msg))

let (find_no_exc : Soc.key -> Soc.tbl -> Soc.t) =
  fun sk soc_tbl -> 
    try find (Lxm.dummy "") sk soc_tbl
    with Errors.Compile_error(_,msg) ->
      print_string msg; 
      flush stdout;
      assert false



let gen_index_list n =
  let rec aux acc i n =
    if i<0 then acc else aux (i::acc) (i-1) n
  in
  aux [] (n-1) n

let _ = assert (gen_index_list 5 = [0;1;2;3;4])

