(* Time-stamp: <modified the 07/02/2013 (at 09:41) by Erwan Jahier> *)

(** Sub module of EvalClock that defines clock-checking utilities.

   We have Two kind of substitutions. 

   - subst1 is used to deal with the clock name binding, to relate node
   parameters and node arguments.

    - subst2 is used to deal with polymorphic clock variables. Indeed,
    constant clock is intrinsically polymorphic. Hence, when clocking
    a contant (e.g., "42"), we return the clock_info Ci_var i, where i
    is a fresh integer. Afterwards, when encoutering an expression
    such as "42+x", where we know the clock of x (clk_x), we produce
    the substitution (i,clk_x).

    XXX Shouldn't I merge the 2 kind of substitutions ?  That would
    mean I need to invent a fake var_info_eff, but it would make the
    whole more homogeous.

    XXX Make it abstract?
*)

type subst1 = (Ident.t * Ident.t) list
type subst2

type subst = subst1 * subst2
(* = (Ident.t * Ident.t) list  * (int * Lic.clock) list *)

val empty_subst : subst


val apply_subst:subst -> Lic.clock -> Lic.clock

(* only apply the part of the subst that deals with clock var *)
val apply_subst2:subst -> Lic.clock -> Lic.clock
val add_subst2 : int -> Lic.clock -> subst2 -> subst2
val add_link2 : int -> int -> subst2 -> subst2

val apply_subst_val_exp : subst -> Lic.val_exp -> Lic.val_exp

(** Raises a Compile_error is the 2 Lic.clock are not unifiable *)
val f : Lxm.t -> subst -> Lic.clock -> Lic.clock -> subst
val list : Lxm.t -> Lic.clock list -> subst -> Lic.clock * subst

val new_clock_var : subst -> subst * Lic.clock

val const_to_val_eff: Lxm.t -> bool -> subst -> Lic.const -> subst * Lic.val_exp  
